"""
Zurich Instruments LabOne Python API Example

Demonstrate how to save and load Zurich Instruments device settings using the
zhinst.utils utility functions save_settings() and load_settings().
"""

# Copyright 2015 Zurich Instruments AG

import time
import os

import zhinst.utils as utils
import zhinst.ziPython as ziPython


def run_example(device_id, settings_file_path=None):
    """
    Run the example: Connect to a Zurich Instruments instrument, save the
    instrument's settings to file, toggle the signal output enable and reload
    the settings file.

    This example demonstrates the use of the zhinst utility functions
    save_settings() and load_settings(). These functions will block until
    saving/loading has finished (i.e., they are synchronous functions). Since
    this is the desired behaviour in most cases, these utility functions are the
    recommended way to save and load settings.

    If an asynchronous interface for saving and loading settings is required,
    please refer to the `example_save_device_settings_expert'. Which
    demonstrates how to directly use the ziDeviceSettings module.

    Arguments:

      device_id (str): The ID of the device to run the example with. For
        example, `dev2006` or `uhf-dev2006`.

      setting_file_path (str, optional): Specify the path where to save the
        settings file.

    Raises:

      RuntimeError: If the device is not connected to the Data Server.
    """
    # Create an instance of the ziDiscovery class.
    d = ziPython.ziDiscovery()

    # Determine the device identifier from it's ID.
    device = d.find(device_id).lower()

    # Get the device's default connectivity properties.
    props = d.get(device)

    # The maximum API level supported by this example.
    apilevel_example = 5
    # The maximum API level supported by the device class, e.g., MF.
    apilevel_device = props['apilevel']
    # Ensure we run the example using a supported API level.
    apilevel = min(apilevel_device, apilevel_example)
    # See the LabOne Programming Manual for an explanation of API levels.

    # Create a connection to a Zurich Instruments Data Server (an API session)
    # using the device's default connectivity properties.
    daq = ziPython.ziDAQServer(props['serveraddress'], props['serverport'], apilevel)

    # Check that the device is visible to the Data Server
    if device not in utils.devices(daq):
        raise RuntimeError("The specified device `%s` is not visible to the Data Server, " % device_id +
                           "please ensure the device is connected by using the LabOne User Interface " +
                           "or ziControl (HF2 Instruments).")

    timestr = time.strftime("%Y%m%d_%H%M")
    filename = timestr + '_settings.xml'  # Change this to the filename you want to save.
    if settings_file_path:
        filename = settings_file_path + os.sep + filename

    toggle_device_setting(daq, device)

    # Save settings
    print "Saving settings..."
    utils.save_settings(daq, device, filename)
    print "Done."

    # Check we actually saved the file
    assert os.path.isfile(filename), "Failed to save settings file '%s'" % filename
    print "Saved file '%s'." % filename

    toggle_device_setting(daq, device)

    # Load settings
    print "Loading settings..."
    utils.load_settings(daq, device, filename)
    print "Done."


def toggle_device_setting(daq, device):
    """
    Toggle a setting on the device: If it's enabled, disable the setting, and
    vice versa.
    """
    path = '/%s/sigouts/0/on' % device
    is_enabled = daq.getInt(path)
    print "Toggling setting '%s'." % path
    daq.setInt(path, not is_enabled)
    daq.sync()
