"""
Zurich Instruments LabOne Python API Example

Demonstrate how to connect to a Zurich Instruments UHF Lock-in Amplifier and
obtain output from the Input PWA and Boxcar using ziDAQServer's poll() command.
"""

# Copyright 2015 Zurich Instruments AG

import re
import time
import warnings
import numpy as np

import zhinst.utils as utils
import zhinst.ziPython as ziPython


def run_example(device_id, do_plot=False):
    """
    Run the example: Connect to a Zurich Instruments UHF Lock-in Amplifier and
    obtain Input PWA and Boxcar data via ziDAQServer's synchronous poll command.

    Requirements:

      UHFLI with UHF-BOX Boxcar Averager Option.

      Hardware configuration: Connect signal output 1 to signal input 1 with a
        BNC cable.

    Arguments:

      device_id (str): The ID of the device to run the example with. For
        example, `dev2006` or `uhf-dev2006`.

      do_plot (bool, optional): Specify whether to plot the boxcar and inputpwa
        output. Default is no plot output.

    Returns:

      sample (dict): A dictionary containing the boxcar sample.

    Raises:

      RuntimeError: If the device is not connected to the Data Server.

      Exception: If the BOX Option is not installed.

      Exception: If an invalid API Level is used.
    """
    # Create an instance of the ziDiscovery class.
    d = ziPython.ziDiscovery()

    # Determine the device identifier from it's ID.
    device = d.find(device_id).lower()

    # Get the device's default connectivity properties.
    props = d.get(device)

    # The maximum API level supported by this example.
    apilevel_example = 5
    # The maximum API level supported by the device class, e.g., MF.
    apilevel_device = props['apilevel']
    # Ensure we run the example using a supported API level.
    apilevel = min(apilevel_device, apilevel_example)
    # See the LabOne Programming Manual for an explanation of API levels.

    # Create a connection to a Zurich Instruments Data Server (an API session)
    # using the device's default connectivity properties.
    daq = ziPython.ziDAQServer(props['serveraddress'], props['serverport'], apilevel)

    # Check that the device is visible to the Data Server
    if device not in utils.devices(daq):
        raise RuntimeError("The specified device `%s` is not visible to the Data Server, " % device_id +
                           "please ensure the device is connected by using the LabOne User Interface " +
                           "or ziControl (HF2 Instruments).")

    # Find out whether the device is an HF2 or a UHF
    devtype = daq.getByte('/%s/features/devtype' % device)
    options = daq.getByte('/%s/features/options' % device)

    if not re.search('UHF', devtype):
        raise Exception("This example can only be ran on UHF Instruments.")

    if not re.search('BOX', options):
        raise Exception("This example can only be ran on a UHF with the BOX option enabled.")

    if daq.getConnectionAPILevel() < 4:
        warnings.warn("ziDAQServer is using API Level 1, it is strongly recommended to use API " +
                      "Level 5 in order to obtain boxcar data with timestamps.")

    # Define configuration and configure the device
    in_c = 0
    out_c = 0
    osc_c = 0
    boxcar_c = 0
    inputpwa_c = 0
    amplitude = 0.5
    frequency = 9.11e6
    windowstart = 75  # boxcar windowstart [degrees]
    windowsize = 3e-9  # boxcar windowsize [seconds]
    if re.match('UHF', devtype) and (not re.search('MF', options)):
        out_mixer_c = 3
    else:
        out_mixer_c = 0

    # create a base configuration: disable all outputs, demods and scopes
    general_setting = [['/%s/demods/*/rate'       % device, 0],
                       ['/%s/demods/*/trigger'    % device, 0],
                       ['/%s/sigouts/*/enables/*' % device, 0],
                       ['/%s/demods/*/enable'     % device, 0],
                       ['/%s/scopes/*/enable'     % device, 0]]
    daq.set(general_setting)

    # now the settings relevant to this experiment
    periods_vals = np.logspace(0, 9, 10, base=2)
    exp_setting = [['/%s/sigins/%d/imp50'           % (device, in_c), 1],
                   ['/%s/sigins/%d/ac'              % (device, in_c), 0],
                   ['/%s/sigins/%d/range'           % (device, in_c), 2*amplitude],
                   ['/%s/demods/%d/oscselect'       % (device, out_mixer_c), osc_c],
                   ['/%s/inputpwas/%d/oscselect'    % (device, inputpwa_c), osc_c],
                   ['/%s/inputpwas/%d/inputselect'  % (device, inputpwa_c), in_c],
                   ['/%s/inputpwas/%d/mode'         % (device, inputpwa_c), 1],
                   ['/%s/inputpwas/%d/shift'        % (device, inputpwa_c), 0.0],
                   ['/%s/inputpwas/%d/harmonic'     % (device, inputpwa_c), 1],
                   ['/%s/inputpwas/%d/enable'       % (device, inputpwa_c), 1],
                   ['/%s/boxcars/%d/oscselect'      % (device, boxcar_c), osc_c],
                   ['/%s/boxcars/%d/inputselect'    % (device, boxcar_c), in_c],
                   ['/%s/boxcars/%d/windowstart'    % (device, boxcar_c), windowstart],
                   ['/%s/boxcars/%d/windowsize'     % (device, boxcar_c), windowsize],
                   ['/%s/boxcars/%d/limitrate'      % (device, boxcar_c), 1e3],
                   ['/%s/boxcars/%d/periods'        % (device, boxcar_c), periods_vals[0]],
                   ['/%s/boxcars/%d/enable'         % (device, boxcar_c), 1],
                   ['/%s/oscs/%d/freq'              % (device, osc_c), frequency],
                   ['/%s/sigouts/%d/on'             % (device, out_c), 1],
                   ['/%s/sigouts/%d/enables/%d'     % (device, out_c, out_mixer_c), 1],
                   ['/%s/sigouts/%d/range'          % (device, out_c), 1],
                   ['/%s/sigouts/%d/amplitudes/%d'  % (device, out_c, out_mixer_c), amplitude]]
    daq.set(exp_setting)

    # Wait for boxcar output to settle
    time.sleep(periods_vals[0]/frequency)

    # Perform a global synchronisation between the device and the data server:
    # Ensure that the settings have taken effect on the device before issuing
    # the poll().
    daq.sync()

    # Get the values that were actually set on the device
    frequency_set = daq.getDouble('/%s/oscs/%d/freq' % (device, osc_c))
    windowstart_set = daq.getDouble('/%s/boxcars/%d/windowstart' % (device, boxcar_c))
    windowsize_set = daq.getDouble('/%s/boxcars/%d/windowsize' % (device, boxcar_c))

    # Subscribe to the nodes we would like to record data from
    boxcar_sample_path = '/%s/boxcars/%d/sample' % (device, boxcar_c)
    boxcar_periods_path = '/%s/boxcars/%d/periods' % (device, boxcar_c)
    inputpwa_wave_path = '/%s/inputpwas/%d/wave' % (device, inputpwa_c)
    daq.subscribe([boxcar_sample_path, boxcar_periods_path, inputpwa_wave_path])
    # We use getAsEvent() to ensure we obtain the first ``periods`` value; if
    # its value didn't change, the server won't report the first value.
    daq.getAsEvent(boxcar_periods_path)

    for periods in periods_vals:
        time.sleep(1.0)
        daq.setInt(boxcar_periods_path, int(periods))

    # Poll the data
    poll_length = 0.1  # [s]
    poll_timeout = 500  # [ms]
    poll_flags = 0
    poll_return_flat_dict = True
    data = daq.poll(poll_length, poll_timeout, poll_flags, poll_return_flat_dict)

    # Unsubscribe from all paths
    daq.unsubscribe('*')

    # Check the dictionary returned by poll contains the subscribed data. The
    # data returned is a dictionary with keys corresponding to the recorded
    # data's path in the node hierarchy
    assert data, "poll returned an empty data dictionary, did you subscribe to any paths?"
    assert boxcar_sample_path in data, "data dictionary has no key '%s'" % boxcar_sample_path
    assert boxcar_periods_path in data, "data dictionary has no key '%s'" % boxcar_periods_path
    assert inputpwa_wave_path in data, "data dictionary has no key '%s'" % inputpwa_wave_path

    sample = data[boxcar_sample_path]

    if daq.getConnectionAPILevel() >= 4:
        # When using API Level 4 (or higher) poll() returns both the 'value' and
        # 'timestamp' of the node. These are two vectors of the same length;
        # which consist of (timestamp,value) pairs.
        boxcar_value = sample['value']
        boxcar_timestamp = sample['timestamp']
        boxcar_periods_value = data[boxcar_periods_path]['value']
        boxcar_periods_timestamp = data[boxcar_periods_path]['timestamp']
    elif daq.getConnectionAPILevel() == 1:
        # When using API Level 1 poll() returns only the 'value' in the node itself.
        boxcar_value = sample
        boxcar_periods_value = data[boxcar_periods_path]
    else:
        raise Exception("Invalid API Level.")

    print "Measured average boxcar amplitude is %.5e V." % np.mean(boxcar_value)

    if do_plot:
        if daq.getConnectionAPILevel() == 4:
            # get the sample rate of the device's ADCs
            clockbase = float(daq.getInt('/%s/clockbase' % device))
            # convert timestamps from ticks to seconds via clockbase
            boxcar_t = (boxcar_timestamp - boxcar_timestamp[0])/clockbase
            boxcar_periods_t = (boxcar_periods_timestamp - boxcar_periods_timestamp[0])/clockbase
            boxcar_periods_t[0] = boxcar_t[0]
        # Create plot
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax1 = fig.add_subplot(111)
        ax2 = ax1.twinx()
        plt.grid(True)
        if daq.getConnectionAPILevel() == 4:
            ax1.plot(boxcar_t, boxcar_value, label='boxcar output')
            plt.xlabel('Time (s)')
            ax2.step(np.append(boxcar_periods_t, boxcar_t[-1]),
                     np.append(boxcar_periods_value, boxcar_periods_value[-1]), '-r', label='Averaging periods')
            ax2.set_yscale("log")
            plt.xlim(min(boxcar_t[0], boxcar_periods_t[0]), max(boxcar_t[-1], boxcar_periods_t[-1]))
            ax2.legend(loc=1)
        else:
            plt.plot(boxcar_value)
            plt.xlabel('Samples: Use API Level 4 for timestamps!')
            plt.xlim(0, len(boxcar_value))
        plt.title('Boxcar output: The effect of averaging\nperiods on the boxcar value.')
        ax1.legend(loc=4)
        ax1.set_ylabel('Boxcar value (V)')
        ax2.set_ylabel('Number of Averaging Periods')
        plt.draw()
        plt.show()

        plt.figure()
        plt.grid(True)
        pwa_wave = data[inputpwa_wave_path][-1]
        pwa_wave['binphase'] = pwa_wave['binphase']*360/(2*np.pi)
        plt.axhline(0, color='k')
        # The inputpwa waveform is stored in 'x', currently 'y' is unused.
        plt.plot(pwa_wave['binphase'], pwa_wave['x'])
        windowsize_set_degrees = 360*frequency_set*windowsize_set
        phase_window = (pwa_wave['binphase'] >= windowstart_set) & \
            (pwa_wave['binphase'] <= windowstart_set + windowsize_set_degrees)
        plt.fill_between(pwa_wave['binphase'], 0, pwa_wave['x'], where=phase_window, alpha=0.5)
        plt.xlim(0, 360)
        title = 'Input PWA waveform, the shaded region shows the portion\n of the waveform the boxcar is integrating.'
        plt.title(title)
        plt.xlabel('Phase (degrees)')
        plt.ylabel('Amplitude (V)')
        plt.draw()
        plt.show()

    return sample
