"""
Zurich Instruments LabOne Python API Example

Demonstrate how to connect to a Zurich Instruments UHF Lock-in Amplifier and use
the PID Advisor to set up an internal PLL control loop using ziDAQServer's
pidAdvisor module.
"""

# Copyright 2015 Zurich Instruments AG

import re
import time
import warnings
import numpy as np

import zhinst.utils as utils
import zhinst.ziPython as ziPython


def run_example(device_id, do_plot=False):
    """
    Run the example: Connect to a Zurich Instruments UHF Lock-in Amplifier and
    obtain optimized P, I, and D parameters for an internal PLL loop using
    ziDAQServer's pidAdvisor module.

    Requirements:

      UHFLI with UHF-PID Quad PID/PLL Controller Option.

      Hardware configuration: Connect signal output 1 to signal input 1 with a
        BNC cable.

    Arguments:

      device_id (str): The ID of the device to run the example with. For
        example, `dev2006` or `uhf-dev2006`.

      do_plot (bool, optional): Specify whether to plot the calculated PLL
        model response. Default is no plot output.

    Returns:

      result (dict): A dictionary containing the PID Advisor result.

    Raises:

      RuntimeError: If the device is not connected to the Data Server.

      Exception: If the PID Option is not installed.

      Exception: If an invalid API Level is used.
    """
    # Create an instance of the ziDiscovery class.
    d = ziPython.ziDiscovery()

    # Determine the device identifier from it's ID.
    device = d.find(device_id).lower()

    # Get the device's default connectivity properties.
    props = d.get(device)

    # The maximum API level supported by this example.
    apilevel_example = 5
    # The maximum API level supported by the device class, e.g., MF.
    apilevel_device = props['apilevel']
    # Ensure we run the example using a supported API level.
    apilevel = min(apilevel_device, apilevel_example)
    # See the LabOne Programming Manual for an explanation of API levels.

    # Create a connection to a Zurich Instruments Data Server (an API session)
    # using the device's default connectivity properties.
    daq = ziPython.ziDAQServer(props['serveraddress'], props['serverport'], apilevel)

    # Check that the device is visible to the Data Server
    if device not in utils.devices(daq):
        raise RuntimeError("The specified device `%s` is not visible to the Data Server, " % device_id +
                           "please ensure the device is connected by using the LabOne User Interface " +
                           "or ziControl (HF2 Instruments).")

    # Find out whether the device is an HF2 or a UHF
    devtype = daq.getByte('/%s/features/devtype' % device)
    options = daq.getByte('/%s/features/options' % device)

    if not re.search('UHF', devtype):
        raise Exception("This example can only be ran on UHF Instruments.")

    if not re.search('PID', options):
        raise Exception("This example can only be ran on a UHF with the PID option enabled.")

    if daq.getConnectionAPILevel() < 4:
        warnings.warn("ziDAQServer is using API Level <4, it is strongly recommended to use API " +
                      "Level >=a in order to obtain pid advisor data.")

    # Define configuration and configure the device

    # pid
    target_bw = 100e3            # Target bandwidth (Hz)
    pid_no = 0                   # PID number
    pid_input = 3                # PID input (3 = Demod phase)
    pid_input_channel = 0        # Demodulator number
    setpoint = 0                 # Phase setpoint
    phase_unwrap = True
    pid_output = 2               # PID output (2 = oscillator frequency)
    pid_output_channel = 0       # number of oscillator controlled by PID
    pid_center_frequency = 10e6  # (Hz)
    pid_limits = 1e6             # (Hz)

    out_c = 0
    osc_c = 1
    out_mixer_c = 4
    amplitude = 1.0

    # create a base configuration: disable all outputs, demods and scopes
    general_setting = [['/%s/demods/*/rate'       % device, False],
                       ['/%s/demods/*/trigger'    % device, False],
                       ['/%s/sigouts/*/enables/*' % device, False],
                       ['/%s/demods/*/enable'     % device, False],
                       ['/%s/scopes/*/enable'     % device, False]]
    daq.set(general_setting)

    # now the settings relevant to this experiment
    exp_setting = [['/%s/pids/%d/input'            % (device, pid_no), pid_input],
                   ['/%s/pids/%d/inputchannel'            % (device, pid_no), pid_input_channel],
                   ['/%s/pids/%d/setpoint'                % (device, pid_no), setpoint],
                   ['/%s/pids/%d/phaseunwrap'             % (device, pid_no), phase_unwrap],
                   ['/%s/pids/%d/output'                  % (device, pid_no), pid_output],
                   ['/%s/pids/%d/outputchannel'           % (device, pid_no), pid_output_channel],
                   ['/%s/pids/%d/center'                  % (device, pid_no), pid_center_frequency],
                   ['/%s/pids/%d/limitlower'              % (device, pid_no), -pid_limits],
                   ['/%s/pids/%d/limitupper'              % (device, pid_no), pid_limits],
                   ['/%s/pids/%d/enable'                  % (device, pid_no), False],
                   ['/%s/oscs/%d/freq'                    % (device, osc_c), pid_center_frequency],
                   ['/%s/sigouts/%d/on'                   % (device, out_c), True],
                   ['/%s/sigouts/%d/enables/%d'           % (device, out_c, out_mixer_c), True],
                   ['/%s/sigouts/%d/range'                % (device, out_c), 1.0],
                   ['/%s/sigouts/%d/amplitudes/%d'        % (device, out_c, out_mixer_c), amplitude]]

    daq.set(exp_setting)

    # Perform a global synchronisation between the device and the data server:
    # Ensure that the settings have taken effect on the device before starting the pidAdvisor.
    daq.sync()

    # set up PID Advisor
    pidAdvisor = daq.pidAdvisor(500)

    # Turn off auto-calc on param change. Enabled
    # auto calculation can be used to automatically
    # update response data based on user input.
    pidAdvisor.set('pidAdvisor/auto', False)
    pidAdvisor.set('pidAdvisor/device', device)
    pidAdvisor.set('pidAdvisor/pid/targetbw', target_bw)

    # PID advising mode (bit coded)
    # bit 0: optimize/tune P
    # bit 1: optimize/tune I
    # bit 2: optimize/tune D
    # Example: mode = 7: Optimize/tune PID
    pidAdvisor.set('pidAdvisor/pid/mode', 7)

    # PID index to use (first PID of device: 0)
    pidAdvisor.set('pidAdvisor/index', pid_no)

    # DUT model
    # source = 1: Lowpass first order
    # source = 2: Lowpass second order
    # source = 3: Resonator frequency
    # source = 4: Internal PLL
    # source = 5: VCO
    # source = 6: Resonator amplitude
    pidAdvisor.set('pidAdvisor/dut/source', 4)

    # IO Delay of the feedback system describing the earliest response
    # for a step change. This parameter does not affect the shape of
    # the DUT transfer function
    pidAdvisor.set('pidAdvisor/dut/delay', 0.0)

    # Other DUT parameters (not required for the internal PLL model)
    # pidAdvisor.set('pidAdvisor/dut/gain', 1.0)
    # pidAdvisor.set('pidAdvisor/dut/bw', 1000)
    # pidAdvisor.set('pidAdvisor/dut/fcenter', 15e6)
    # pidAdvisor.set('pidAdvisor/dut/damping', 0.1)
    # pidAdvisor.set('pidAdvisor/dut/q', 10e3)

    # Start values for the PID optimization. Zero
    # values will initate a guess. Other values can be
    # used as hints for the optimization process.
    pidAdvisor.set('pidAdvisor/pid/p', 0)
    pidAdvisor.set('pidAdvisor/pid/i', 0)
    pidAdvisor.set('pidAdvisor/pid/d', 0)
    pidAdvisor.set('pidAdvisor/calculate', False)

    # Start the module thread
    pidAdvisor.execute()

    # Advise
    pidAdvisor.set('pidAdvisor/calculate', True)
    print 'Starting advising. Optimization process may run up to a minute...'
    reply = pidAdvisor.get('pidAdvisor/calculate')

    t_start = time.time()
    t_timeout = t_start + 25
    while reply['calculate'][0] is True:
        reply = pidAdvisor.get('pidAdvisor/calculate')
        if time.time() > t_timeout:
            pidAdvisor.finish()
            raise Exception("PID advising failed due to timeout.")

    pidAdvisor.set('pidAdvisor/calculate', True)
    print "Advice took %0.1f s" % (time.time() - t_start)

    # Get all calculated parameters
    result = pidAdvisor.get('pidAdvisor/*')
    # Check that the dictionary returned by poll contains the data that are needed
    assert result, "pidAdvisor returned an empty data dictionary?"
    assert 'pid' in result, "data dictionary has no key 'pid'"
    assert 'step' in result, "data dictionary has no key 'step'"
    assert 'bode' in result, "data dictionary has no key 'bode'"

    if result is not None:
        # Now copy the values from the PID Advisor to the PID and enable the PID
        p_adv = result['pid']['p'][0]
        i_adv = result['pid']['i'][0]
        d_adv = result['pid']['d'][0]
        dlimittimeconstant_adv = result['pid']['dlimittimeconstant'][0]
        rate_adv = result['pid']['rate'][0]
        bw_adv = result['bw'][0]

        advisor_result_setting = [['/%s/pids/%d/p'                  % (device, pid_no), p_adv],
                                  ['/%s/pids/%d/i'                  % (device, pid_no), i_adv],
                                  ['/%s/pids/%d/d'                  % (device, pid_no), d_adv],
                                  ['/%s/pids/%d/dlimittimeconstant' % (device, pid_no), dlimittimeconstant_adv],
                                  ['/%s/pids/%d/rate'               % (device, pid_no), rate_adv],
                                  ['/%s/pids/%d/enable'             % (device, pid_no), True]]

        daq.set(advisor_result_setting)

    if do_plot:
        import matplotlib.pyplot as plt
        plt.close('all')
        bode_complex_data = result['bode'][0]['x'] + 1j * result['bode'][0]['x']
        bode_grid = result['bode'][0]['grid']

        step_x = result['step'][0]['x']
        step_grid = result['step'][0]['grid']

        _, ax_list = plt.subplots(2, 1)
        ax_list[0].semilogx(bode_grid, 20*np.log10(np.abs(bode_complex_data)))
        ax_list[0].grid(True)
        ax_list[0].set_title("Calculated model response for internal PLL with " +
                             "P = %0.0f, I = %0.0f,\n""D = %0.5f and bandwidth %0.0f kHz"
                             % (p_adv, i_adv, d_adv, bw_adv* 1e-3))
        ax_list[0].set_ylabel('Bode Gain (dB)')

        ax_list[1].semilogx(bode_grid, np.angle(bode_complex_data)/np.pi*180)
        ax_list[1].grid(True)
        ax_list[1].set_xlabel('Frequency (Hz)')
        ax_list[1].set_ylabel('Bode Phase (deg)')

        _, ax = plt.subplots(1, 1)
        ax.plot(step_grid*1e6, step_x)
        ax.grid(True)
        ax.set_title("Calculated model response for internal PLL with " +
                     "P = %0.0f, I = %0.0f,\nD = %0.5f and bandwidth %0.0f kHz"
                     % (p_adv, i_adv, d_adv, bw_adv* 1e-3))
        ax.set_xlabel('Time (us)')
        ax.set_ylabel('Step Response')

        plt.draw()
        plt.show()

    return result
